/*
 *   Copyright (c) 1998 T. Kamei (kamei@jsdlab.co.jp)
 *
 *   Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted provided
 * that the above copyright notice and this permission notice appear
 * in all copies of the software and related documentation.
 *
 *                          NO WARRANTY
 *
 *   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES;
 * WITHOUT EVEN THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
 * FOR A PARTICULAR PURPOSE.
 */

#include <windows.h>
#include <stdio.h>
#include "cmdline.h"

void
mapf::close ()
{
	if (mf_base)
		UnmapViewOfFile (mf_base);
	if (mf_hm)
		CloseHandle (mf_hm);
	if (mf_hf != INVALID_HANDLE_VALUE)
		CloseHandle (mf_hf);
	init ();
}

int
mapf::open (const char *path, int mode)
{
	close ();
	mf_hf = CreateFile (path, GENERIC_READ, FILE_SHARE_READ, 0,
		OPEN_EXISTING, mode, 0);
	if (mf_hf == INVALID_HANDLE_VALUE)
		return 0;
	
	mf_size = GetFileSize (mf_hf, 0);
	if (mf_size)
	{
		mf_hm = CreateFileMapping (mf_hf, 0, PAGE_READONLY, 0, 0, 0);
		if (!mf_hm)
			return 0;
		
		mf_base = MapViewOfFile (mf_hm, FILE_MAP_READ, 0, 0, 0);
		if (!mf_base)
			return 0;
	}
	return 1;
}

void
cmdline::discard ()
{
	for (int i = 0; i < cl_ac; i++)
	{
		if (cl_av[i])
			free (cl_av[i]);
	}
	if (cl_av)
		free (cl_av);
	cl_ac = 0;
	cl_av = 0;
	cl_max = 0;
}

char *
cmdline::copyarg (const u_char *s, const u_char *se, int dq)
{
	int l = se - s - dq;
	if (!l)
		return "";
	char *d0 = (char *)malloc (l + 1);
	if (!d0)
		return 0;
	char *d = d0; // 2001.6.17: bcc32 ΍ S.Noda 
	for (; s < se; s++)
	{
		if (*s != '"')
			*d++ = *s;
	}
	*d = 0;
	return d0;
}

int
cmdline::parse (const char *cmdline, int l, int resp_ok)
{
	if (!cmdline)
		return 0;
	const u_char *cp = (const u_char *)cmdline;
	const u_char *ce = cp + l;
	const u_char *c0 = 0;
	int dq = 0;
	for (;; cp++)
	{
		int c = cp == ce ? EOF : *cp;
		switch (c)
		{
		case ' ':
		case '\t':
			if (dq & 1)
				goto normal_char;
			/* fall thru... */
		case EOF:
		case '\r':
		case '\n':
			if (c0)
			{
				char *arg = copyarg (c0, cp, dq);
				if (!arg)
					return ERROR_ENOUGH_MEMORY;
				if (resp_ok && *arg == '@')
				{
					mapf mf;
					int e = mf.open (arg + 1);
					free (arg);
					if (!e)
						return ERROR_RESPONSE_READ;
					e = parse ((const char *)mf.base (), mf.size (), 0);
					if (e)
						return e;
				}
				else if (*arg)
				{
					if (cl_ac == cl_max)
					{
						cl_max += 32;
						char **x = (char **)realloc (cl_av, sizeof *x * cl_max);
						if (!x)
						{
							free (arg);
							return ERROR_ENOUGH_MEMORY;
						}
						cl_av = x;
					}
					cl_av[cl_ac++] = arg;
				}
				dq = 0;
				c0 = 0;
			}
			if (c == EOF)
				return 0;
			break;
			
		case '"':
			dq++;
			/* fall thru... */
		default:
normal_char:
			if (!c0)
				c0 = cp;
			break;
		}
	}
}
