/*
 *   Copyright (c) 2000-2002 S.Noda <sakae_noda@hotmail.com>
 *
 *   Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted provided
 * that the above copyright notice and this permission notice appear
 * in all copies of the software and related documentation.
 *
 *                          NO WARRANTY
 *
 *   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES;
 * WITHOUT EVEN THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
 * FOR A PARTICULAR PURPOSE.
 */

#include <crtdbg.h>
#include <windows.h>
#include <stdio.h>
#include "cmdline.h"
#include "zipdll.h"

#ifndef _CRT_WARN
#define _CRT_WARN           0
#define _CRT_ERROR          1
#define _CRT_ASSERT         2
#define _CRT_ERRCNT         3
#endif

#ifndef _RPTF0
#define _RPTF0(rptno, msg)
#define _RPTF1(rptno, msg, arg1)
#define _RPTF2(rptno, msg, arg1, arg2)
#define _RPTF3(rptno, msg, arg1, arg2, arg3)
#define _RPTF4(rptno, msg, arg1, arg2, arg3, arg4)
#endif

#define Z_DLL_NAME "ZIP32.DLL\0"

static char *password = NULL;
static char *sfxdat = NULL;
static int FNV_max = 0;

static int WINAPI ZpPrnt(LPSTR buf, unsigned long size)
{
	printf("%s", buf);
	return (unsigned int)size;
}

static int WINAPI ZpPassword(LPSTR p, int n, LPCSTR m, LPCSTR name)
{
	if (!password)
	{
		fprintf(stderr, "%s", m);
		scanf("%s", p);
	}
	else
		strcpy(p, password);
	return 0;
}

static int WINAPI ZpService(LPCSTR filename, unsigned long size)
{
	_RPTF2(_CRT_WARN, "<ZpService: %s(%d)>\n", filename, size);
	return FALSE;
}

static int WINAPI ZpComment(LPSTR buf)
{
	fprintf(stderr, "enter new zip file comment: ");
	scanf("%s", buf);
	return TRUE;
}

int ZpMain(ZCL CL, LPZPOPT lpOPT)
{
	int ret;
	
	HANDLE hUF = GlobalAlloc(GPTR, (DWORD)sizeof(ZIPUSERFUNCTIONS));
	if (!hUF)
	{
		return -11;
	}
	LPZIPUSERFUNCTIONS lpUserFunc = (LPZIPUSERFUNCTIONS)GlobalLock(hUF);
	if (!lpUserFunc)
	{
		GlobalFree(hUF);
		return -12;
	}
	
	lpUserFunc->print = ZpPrnt;
	lpUserFunc->comment = ZpComment;
	lpUserFunc->password = ZpPassword;
	lpUserFunc->ServiceApplication = ZpService;
	
	HINSTANCE hZipDll = LoadLibrary(Z_DLL_NAME);
	if (!hZipDll)
	{
		GlobalUnlock(hUF);
		GlobalFree(hUF);
		return -13;
	}
	ZP_VERSION fpZpVersion = (ZP_VERSION)GetProcAddress(hZipDll, "ZpVersion");
	ZP_INIT fpZpInit = (ZP_INIT)GetProcAddress(hZipDll, "ZpInit");
	ZP_SETOPTIONS fpZpSetOptions = (ZP_SETOPTIONS)GetProcAddress(hZipDll, "ZpSetOptions");
	ZP_ARCHIVE fpZpArchive = (ZP_ARCHIVE)GetProcAddress(hZipDll, "ZpArchive");
	if (!fpZpVersion || !fpZpInit || !fpZpSetOptions || !fpZpArchive)
	{
		GlobalUnlock(hUF);
		GlobalFree(hUF);
		return -14;
	}
	
	ZpVer ver;
	(*fpZpVersion)(&ver);
	_RPTF1(_CRT_WARN, "<DLL Date: %s>\n", ver.date);
	_RPTF3(_CRT_WARN, "<DLL Version: %d.%d%d>\n", ver.windll.major, ver.windll.minor, ver.windll.patchlevel);
	if (ver.windll.major == 2 && ver.windll.minor >= 3)
	{
		if ((*fpZpInit)(lpUserFunc))
		{
			(*fpZpSetOptions)(lpOPT);
			ret = (*fpZpArchive)(CL);
		} else ret = -16;
	} else ret = -15;
	
	GlobalUnlock(hUF);
	GlobalFree(hUF);
	FreeLibrary(hZipDll);
	return ret; // if (ret != 0) then "Error"
	
}

int ZpMakeSFX(LPZCL lpCL, LPZPOPT lpOPT)
{
	static char dat[PATH_MAX], arc[PATH_MAX], sfx[PATH_MAX];
	FILE *datfp, *arcfp, *sfxfp;
	char buf[1024];
	int in, out;
	
	// dat
	if (!SearchPath(NULL, sfxdat, NULL, sizeof(dat), dat, NULL))
	{
		fcloseall();
		return -21;
	}
	datfp = fopen(dat, "rb");
	if (!datfp)
	{
		fcloseall();
		return -22;
	}
	// arc
	if (lpOPT->szRootDir)
		SetCurrentDirectory(lpOPT->szRootDir);
	strcpy(arc, lpCL->lpszZipFN);
	arcfp = fopen(arc, "rb");
	if (!arcfp)
	{
		strcat(arc, ".zip");
		arcfp = fopen(arc, "rb");
		if (!arcfp)
		{
			fcloseall();
			return -23;
		}
	}
	// sfx
	strcpy(sfx, lpCL->lpszZipFN);
	if (!(strlen(sfx) >= 4 && stricmp(sfx + strlen(sfx) -4 , ".zip") == 0))
		strcat(sfx, ".exe");
	else
		strcpy(sfx + strlen(sfx) - 4, ".exe");
	// ㏑h~
	if (GetFileAttributes(sfx) != 0xFFFFFFFF)
	{
		fcloseall();
		return -24;
	}
	sfxfp = fopen(sfx, "wb");
	if(!sfxfp)
	{
		fcloseall();
		return -25;
	}
	// dat->sfx
	while ((in = fread(buf, 1, 1024, datfp)) > 0)
	{
		out = fwrite(buf, 1, in, sfxfp);
		if (in != out)
		{
			fcloseall();
			return -26;
		}
	}
	// arc->sfx
	while ((in = fread(buf, 1, 1024, arcfp)) > 0)
	{
		out = fwrite(buf, 1, in, sfxfp);
		if (in != out)
		{
			fcloseall();
			return -27;
		}
	}
	fcloseall();
	remove(arc);
	lpCL->lpszZipFN = sfx;
	lpOPT->fOffsets = TRUE;
	return 0;
}

static void ZpDiscardCmdLine(LPZCL lpCL, LPZPOPT lpOPT)
{
	sfxdat = NULL;
	lpOPT->szTempDir = NULL;
	password = NULL;
	lpOPT->Date = NULL;
	lpOPT->szRootDir = NULL;
	lpCL->lpszZipFN = NULL;
	for (int i = 0; i < lpCL->argc; i++)
		lpCL->FNV[i] = NULL;
	
	if (lpCL->FNV)
		free(lpCL->FNV);
	lpCL->argc = 0;
	lpCL->FNV = NULL;
	FNV_max = 0;
}

static int ZpParseCmdLine(LPZCL lpCL, LPZPOPT lpOPT, int ac, char **av)
{
	for (int i = 0; i < ac; i++)
	{
		if (*av[i] == '-')
		{
			char *opt = av[i] + 1;
			if (strncmp(opt, "-sfx", 4) == 0)
			{
				char *p = opt + 4;
				if (*p == '\0')
					sfxdat = "sfx32gui.dat";
				else if (*p == '=')
					sfxdat = p + 1;
				else
					return -31;
			}
			else
			{
				while (*opt)
				{
					switch (*opt)
					{
					case 'b':
						lpOPT->fTemp = TRUE;
						if (++i >= ac)
							return -32;
						lpOPT->szTempDir = av[i];
						break;
					case 'e':
						lpOPT->fEncrypt = TRUE;
						break;
					case 'P':
						lpOPT->fEncrypt = TRUE;
						if (++i >= ac)
							return -33;
						password = av[i];
						break;
					case 'S':
						lpOPT->fSystem = TRUE;
						break;
					case '$':
						lpOPT->fVolume = TRUE;
						break;
					case 'X':
						lpOPT->fExtra = TRUE;
						break;
					case 'D':
						lpOPT->fNoDirEntries = TRUE;
						break;
					case 't':
						if (*(opt + 1) == 't')
						{
							lpOPT->fExcludeDate = TRUE;
							opt++;
						}
						else
							lpOPT->fIncludeDate = TRUE;
						if (++i >= ac)
							return -34;
						lpOPT->Date = av[i];
						break;
					case 'v':
						lpOPT->fVerbose = TRUE;
						break;
					case 'q':
						lpOPT->fQuiet = TRUE;
						break;
					case 'l':
						if (*(opt + 1) == 'l')
						{
							lpOPT->fCRLF_LF = TRUE;
							opt++;
						}
						else
							lpOPT->fLF_CRLF = TRUE;
						break;
					case 'j':
						lpOPT->fJunkDir = TRUE;
						break;
					case 'g':
						lpOPT->fGrow = TRUE;
						break;
					case 'k':
						lpOPT->fForce = TRUE;
						break;
					case 'm':
						lpOPT->fMove = TRUE;
						break;
					case 'd':
						lpOPT->fDeleteEntries = TRUE;
						break;
					case 'u':
						lpOPT->fUpdate = TRUE;
						break;
					case 'f':
						lpOPT->fFreshen = TRUE;
						break;
					case 'J':
						lpOPT->fJunkSFX = TRUE;
						break;
					case 'o':
						lpOPT->fLatestTime = TRUE;
						break;
					case 'z':
						lpOPT->fComment = TRUE;
						break;
					case 'A':
						lpOPT->fOffsets = TRUE;
						break;
					case '!':
						lpOPT->fPrivilege = TRUE;
						break;
					case 'r':
						lpOPT->fRecurse = 1;
						break;
					case 'R':
						lpOPT->fRecurse = 2;
						break;
					case 'F':
						if (*(opt + 1) == 'F')
						{
							lpOPT->fRepair = 2;
							opt++;
						}
						else
							lpOPT->fRepair = 1;
						break;
					default:
						if (isdigit(*opt))
							lpOPT->fLevel = *opt;
						else
							return -35;
						break;
					}
					opt++;
				}
			}
		}
		else
		{
			if (*(av[i] + strlen(av[i]) - 1) == '\\')
			{
				*(av[i] + strlen(av[i]) - 1) = '\0';
				lpOPT->szRootDir = av[i];
			}
			else if (lpCL->lpszZipFN == NULL)
			{
				lpCL->lpszZipFN = av[i];
			}
			else
			{
				if (lpCL->argc == FNV_max)
				{
					FNV_max += 32;
					char **x = (char **)realloc(lpCL->FNV, sizeof *x * FNV_max);
					if (!x)
						return -36;
					lpCL->FNV = x;
				}
				lpCL->FNV[lpCL->argc++] = av[i];
			}
		}
	}
	return 0; // if (ret != 0) then "Error"
}

int ZpMain(LPTSTR line)
{
	int ret;
	cmdline cl;
	
	ZCL CL;
	ZPOPT OPT;
	memset(&CL,'\0',sizeof(ZCL));
	memset(&OPT,'\0',sizeof(ZPOPT));
	
	ret = cl.parse(line, 1);
	if (ret)
	{
		switch (ret)
		{
		case ERROR_ENOUGH_MEMORY:
			fprintf(stderr, "O[osĂ܂B");
			break;
		case ERROR_RESPONSE_READ:
			fprintf(stderr, "X|Xt@C̓ǂݎɎs܂B");
			break;
		default:
			break;
		}
		return -41;
	}
	ret = ZpParseCmdLine(&CL, &OPT, cl.argc() - 1, cl.argv() + 1);
	if (ret)
	{
		switch (ret)
		{
		case -31:
			fprintf(stderr, "--sfxIvV̉͂Ɏs܂B");
			break;
		case -32:
			fprintf(stderr, "ƃfBNgwĂ܂B");
			break;
		case -33:
			fprintf(stderr, "pX[hwĂ܂B");
			break;
		case -34:
			fprintf(stderr, "twĂ܂B");
			break;
		case -35:
			fprintf(stderr, "m̃IvVɑ܂B");
			break;
		case -36:
			fprintf(stderr, "realloc()Ɏs܂B");
			break;
		default:
			break;
		}
		ZpDiscardCmdLine(&CL, &OPT);
		return -42;
	}
	if (!CL.lpszZipFN)
	{
		fprintf(stderr, "Usage: [-<options>...] <archive_file_name>[.zip]\n\t\t\t [<directory_name>\\] [<filespec>...]");
		ZpDiscardCmdLine(&CL, &OPT);
		return -43;
	}
	
	if (sfxdat)
	{
		ret = ZpMakeSFX(&CL, &OPT);
		if (ret)
		{
			switch (ret)
			{
			case -21:
				fprintf(stderr, "ZIPȉ𓀕̌Ɏs܂B");
				break;
			case -22:
				fprintf(stderr, "ZIPȉ𓀕̃I[vɎs܂B");
				break;
			case -23:
				fprintf(stderr, "ZIPɂ̃I[vɎs܂B");
				break;
			case -24:
				fprintf(stderr, "ZIPȉ𓀏ɂłɑ݂Ă܂B");
				break;
			case -25:
				fprintf(stderr, "ZIPȉ𓀏ɂ̃I[vɎs܂B");
				break;
			case -26:
				fprintf(stderr, "ZIPȉ𓀕̏oɎs܂B");
				break;
			case -27:
				fprintf(stderr, "ZIPɂ̏oɎs܂B");
				break;
			default:
				break;
			}
			ZpDiscardCmdLine(&CL, &OPT);
			return -44;
		}
	}
	ret = ZpMain(CL, &OPT);
	switch (ret)
	{
	case -11:
		fprintf(stderr, "GlobalAlloc()Ɏs܂B");
		break;
	case -12:
		fprintf(stderr, "GlobalLock()Ɏs܂B");
		break;
	case -13:
		fprintf(stderr, "ZIP32.DLL̃[hɎs܂B");
		break;
	case -14:
		fprintf(stderr, "ΉZIP32.DLLłB");
		break;
	case -15:
		fprintf(stderr, "Ή̃o[WZIP32.DLLłB");
		break;
	case -16:
		fprintf(stderr, "ZIP32.DLL̏Ɏs܂B");
		break;
	default:
		break;
	}
	ZpDiscardCmdLine(&CL, &OPT);
	
	return ret; // if (ret != 0) then "Error"
}
